from Thuban import _

from Thuban.Model.layer import SHAPETYPE_POLYGON, SHAPETYPE_ARC, \
     SHAPETYPE_POINT
from Thuban.Model.label import ALIGN_CENTER, ALIGN_TOP, ALIGN_BOTTOM, \
     ALIGN_LEFT, ALIGN_RIGHT, ALIGN_BASELINE

from Thuban.Model.classification import Classification
from Thuban.Model.color import Color

import gd

def color2gdcolor(img, color):
    idxcolor = (color.red * 255, color.green * 255, color.blue * 255)
    return img.colorResolve(idxcolor)

def point_int(p):
    return (int(p[0]), int(p[1]))

class GDRenderer:
    """Class to render a map onto a ImageDraw"""

    honor_visibility = 1

    def __init__(self, img, scale, offset, resolution = 72.0,
                 honor_visibility = None):
        """Inititalize the renderer.

        dc -- the wxPython DC to render on
        scale, offset -- the scale factor and translation to convert
                between projected coordinates and the DC coordinates

        resolution -- the assumed resolution of the DC. Used to convert
                absolute lengths like font sizes to DC coordinates

        honor_visibility -- boolean. If true, honor the visibility flag
                of the layers, otherwise draw all layers. If None, use
                the renderer's default.
        """
        # resolution in pixel/inch

        assert scale > 0

        self.img = img
        self.scale = scale
        self.offset = offset
        if honor_visibility is not None:
            self.honor_visibility = honor_visibility
        # store the resolution in pixel/point because it's more useful
        # later.
        self.resolution = resolution / 72.0

        self.transparent_color = self.img.colorAllocate((255, 255, 255))
        self.img.colorTransparent(self.transparent_color)

        tbrush = gd.image((1, 1))
        tcolor = tbrush.colorAllocate((255, 255, 255))
        tbrush.colorTransparent(tcolor)
        tbrush.setPixel((0, 0), tcolor)
        self.transparent_brush = tbrush
        
    def render_map(self, map):
        self.map = map
        for layer in map.Layers():
            # if honor_visibility is true, only draw visible layers,
            # otherwise draw all layers
            if not self.honor_visibility or layer.Visible():
                self.draw_shape_layer(layer)
        self.draw_label_layer(map.LabelLayer())

    def draw_shape_layer(self, layer):
        scale = self.scale
        offx, offy = self.offset

        map_proj = self.map.projection
        layer_proj = layer.projection

        shapetype = layer.ShapeType()

        brush = None
        pen   = None

        old_prop = None
        lc = layer.GetClassification()
        field = lc.GetField()


        for i in self.layer_ids(layer):
            value = None

            if field is not None:
                try:
                    record = layer.table.read_record(i)
                    if record is not None:
                        value = record[field]
                except:
                    pass

                #
                # if the above statements fail 'value' should
                # be null, at which point this call will
                # at least retreive the NullData
                #

                group = lc.FindGroup(value)

                #prop = lc.GetProperties(value)
            else:
                group = lc.GetDefaultGroup()


            if not group.IsVisible():
                continue

            prop = group.GetProperties()

            # don't recreate new objects if they are the same as before
            if prop != old_prop:
                old_prop = prop

                if shapetype == SHAPETYPE_ARC:
                    fill = Color.Transparent
                else:
                    fill = prop.GetFill()

                self.set_fill_color(fill)
                self.set_pen(prop.GetLineColor(), prop.GetLineWidth())
            
            if shapetype == SHAPETYPE_POINT:
                self.draw_point_shape(layer, i)
            elif shapetype == SHAPETYPE_ARC:
                self.draw_arc_shape(layer, i)
            else:
                self.draw_polygon_shape(layer, i)
            
    def layer_ids(self, layer):
        """Return the shape ids of the given layer that have to be drawn.

        The default implementation simply returns all ids in the layer.
        Override in derived classes to be more precise.
        """
        return range(layer.NumShapes())

    def draw_polygon_shape(self, layer, index):
        pp = self.projected_points(layer, index)[:-1]
        print "poly:", pp
        
        if len(pp) < 4: return # ignore Null Shapes which have no points

        if self.fill_color is not self.transparent_color:
            self.img.filledPolygon(pp, self.fill_color)
        if self.pen is not self.transparent_color:
            self.img.polygon(pp, self.pen)

    def projected_points(self, layer, index):
        proj = self.map.projection
        if proj is not None:
            forward = proj.Forward
        else:
            forward = None
        proj = layer.projection
        if proj is not None:
            inverse = proj.Inverse
        else:
            inverse = None
        shape = layer.Shape(index)
        points = []
        scale = self.scale
        offx, offy = self.offset
        for x, y in shape.Points():
            if inverse:
                x, y = inverse(x, y)
            if forward:
                x, y = forward(x, y)
            points.append((int(x * scale + offx), int(-y * scale + offy)))
        return points

    def draw_arc_shape(self, layer, index):
        pp = self.projected_points(layer, index)
        print "arc:", pp
        
        if len(pp) < 2: return

        for i in range(len(pp) - 1):
            self.img.line(pp[i], pp[i + 1], self.pen)

    def draw_point_shape(self, layer, index):
        pp = self.projected_points(layer, index)

        if len(pp) == 0: return # ignore Null Shapes which have no points

        p = pp[0]
        radius = self.resolution * 5
        extent = (2*radius, 2*radius)
        if self.fill_color is not self.transparent_color:
            self.img.filledEllipse(p, extent, self.fill_color)
        print "point:", p, radius
        self.img.arc(p, extent, 0, 360, self.pen)

    def draw_label_layer(self, layer):
        scale = self.scale
        offx, offy = self.offset

        map_proj = self.map.projection
        if map_proj is not None:
            forward = map_proj.Forward
        else:
            forward = None

        for label in layer.Labels():
            x = label.x
            y = label.y
            text = label.text
            if forward:
                x, y = forward(x, y)
            x = x * scale + offx
            y = -y * scale + offy
            width, height = self.img.get_bounding_rect(self.FONT_PATH, text)
            if label.halign == ALIGN_LEFT:
                # nothing to be done
                pass
            elif label.halign == ALIGN_RIGHT:
                x = x - width
            elif label.halign == ALIGN_CENTER:
                x = x - width/2
            if label.valign == ALIGN_TOP:
                # nothing to be done
                pass
            elif label.valign == ALIGN_BOTTOM:
                y = y - height
            elif label.valign == ALIGN_CENTER:
                y = y - height/2
            self.img.string_ft(self.FONT_PATH, self.resolution * 10, 0,
                               (x, y), text, gdBrushed)
            
    def make_brush(self, color, width = 1):
        if color is Color.Transparent:
            return self.transparent_brush
        else:
            brush = gd.image((width, width))
            gdcolor = color2gdcolor(brush, color)
            brush.filledEllipse((width / 2, width / 2), (width, width),
                                gdcolor)
            return brush
        
    def set_fill_color(self, color):
        if color is Color.Transparent:
            self.fill_color = self.transparent_color
        else:
            self.fill_color = color2gdcolor(self.img, color)

    def set_pen(self, color, width):
        if width == 1:
            if color is Color.Transparent:
                self.pen = self.transparent_color
            else:
                self.pen = color2gdcolor(self.img, color)
        else:
            self.pen = gd.gdBrushed
            self.img.setBrush(self.make_brush(color, width))
